
""" 
	we want m = k1 d k2
	m2^t = k2^-1 d k1^-1
	m2^t m = k2^-1 d^2 k2
	so the columns of k2^-1 are the eigenvectors of m^t m
	and d = sqrt of eigenvalues
	m =  a b
		 c d
	m^t = a c
		  b d
	m^t m = a^2+c^2 ab+cd ab+cd b^2+d^2
	e^2 - (tr) e + det = 0
	A B x = ex
	B C y   ey
	(A-e)x + By = 0, Bx + (C-e)y = 0
	either [ -B (A-e) ] or [ -(C-e) B ] normalized
"""

import math
def cartan(m):
	a = m[0][0]
	b = m[0][1]
	c = m[1][0]
	d = m[1][1]
	A = a*a+c*c
	B = a*b+c*d
	C = b*b+d*d
	tr = A+C
	det = A*C-B*B
	disc = math.sqrt(tr*tr-4*det)
	e1 = (tr + disc)/2.0 # larger
	e2 = (tr - disc)/2.0
	aB = abs(B)
	aA = abs(A-e1)
	aC = abs(C-e1)
	if aB > aA:
		s = aA/aB
		r1 = aB*math.sqrt(1+s*s)
	else: # aA >= aB 
		if aA == 0:
			r1 = 0
		else:
			s = aB/aA
			r1 = aA*math.sqrt(1+s*s)
	if aB > aC:
		s = aC/aB
		r2 = aB*math.sqrt(1+s*s)
	else: # aC > aB
		if aC == 0:
			r2 = 0
		else:
			s = aB/aC
			r2 = aC*math.sqrt(1+s*s)

	if r1 > r2:
		x = -B/r1; y = (A-e1)/r1
	else:
		print "*"
		x = -(C-e1)/r2; y = B/r2
	# so now [x y] = eigenvector for e1
	# M | x -y  = x -y | e1
	#	| y  x    y  x | e2
	# k2 =  x y
	#      -y x
	print [x, y]
	print [ e1, e2 ]
	return [ x, y, math.sqrt(e1), math.sqrt(e2) ] 

# ==========================================================

def mul(a, b):
	return [
		[ a[0][0]*b[0][0]+a[0][1]*b[1][0],
		  a[0][0]*b[0][1]+a[0][1]*b[1][1] ],
		[ a[1][0]*b[0][0]+a[1][1]*b[1][0],
		  a[1][0]*b[0][1]+a[1][1]*b[1][1] ],
	]

m = [
	[ 1, 0 ],
	[ 0, 2 ]
]

c = cartan(m)

