#!/usr/bin/python

from Path import Path

class proc:
	# b = array of bytes
	def __init__(self, b ):
		self.b = b
		self.i = 0

	def read(self):
		c = self.b[self.i]
		# print "i, c = " + str(self.i) + ", " + str(ord(str(c)))
		self.i += 1
		return(c)
		
"""

Character descriptions occur in both
the subroutines and the character strings of a Type 1 font.
They are part of the binary segment.  After this segment is decrypted,
the character strings are bared - but are revealed as still encrypted  
byte arrays.  These must be in turn decrypted by a slight variant
of the original decryption (different initial value of R)
and each character description is then exposed as a 
sequence of data and commands, in a condensed kind
of PostScript drawing language which uses, as PS does,
an operator opstack and a process opstack.  
The sequence is condensed so as not to take
up much space.  The data involved are integers, usually not very large ones,
and the commands are for the most part relative drawing commands.  
We shall see in a moment how to read this command sequence.
Here is the command list, by category:

Starting and finishing:
	endchar
	hsbw
	seac
	sbw
Path Construction:
	closepath
	hlineto
	hmoveto
	hvcurveto
	rlineto
	rmoveto
	rrcurveto
	vhcurveto
	vlineto
	vmoveto
Hints:
	dotsection
	hstem
	hstem3
	vstem
	vstem3
Arithmetic:
	div
Subroutines:
	callothersubr
	callsubr
	pop
	return
	setcurrentpoint

The interpreter has an operand opstack, which holds integers.
Arguments to an operator sit on the opstack and are removed by it.

In addition, each character description is begun by
some number of meaningless bytes.  By default the number of these 
is 4, but it can be set in the /Private dictionary of the binary segment	by defining /lenIV.

Data (coordinates) of a character's path are
recorded in integers, but fractions are allowed by using
the div operator.  Thus 1/2 is 1 2 div.

Integers are stored in a special way, taking up to 5 bytes:
A byte b on the opstack in the range [32, 256) means an integer n.
Integers are distinguished from commands by the first byte b.  If in
the range [0, 32) it begins an operator; if in [32,256) it begins
an integer.
	if b is in [32, 247), n = b - 139: n in (-108, +108)
	if b is in [247, 251), let w = next byte; n = (b-247)*256 + w + 108:
		n in [108, 1132)
	if b is in [251, 255), w = next byte; n = -(b-251)*256 - w - 108
		n in (-1132, -108]
	if b is in 55, n = next 4 bytes b[i] in big-endian form
		n = b[0]*256^3 + b[1]*256^2 + b[2]*256 + b[3]

-----------------------------------------------------------------

Operators take up 1 or 2 bytes.  The ones which take up two bytes have 12
as the first.  They take their arguements from the operator opstack, but very often
from the bottom of the opstack.  In the follwoing list, 
those which take arguments from the bottom of the opstack are #;
if they clear the opstack we have # ... #; - = empty opstack as argument, but not cleared.

	b in [0, 32) but not 12: n = b
	b = 12 (escape): the next byte w = a command.; see list below.

			 1: # y dy hstem #
			 3: # x dx vstem #
			 4: # dy vmoveto #
				0 dy rmoveto
			 5: # dx dy rlineto #
				dx dy rlineto
			 6: # dx hlineto #
				x 0 rlineto
			 7: # dy vlineto #
				0 dy rlineto
			 8: # dx1 dy1 dx2 dy2 dx3 dy3 rrcurveto #
				dx1 dy1 (dx1+dx2) (dy1+dy2) (dx1+dx2+dx3) (dy1+dy2+dy3) rcurveto
			 9: - closepath #
				closepath
			10: subr# callsubr -
			11: - return -
			12: escape
				read next byte to get command
			13: # sbx wx hsbw #
				same as 
					sbx 0 wx 0 sbw
			14: - endchar #
			21: # dx dy rmoveto #
				dx dy rmoveto
			22: # dx hmoveto #
				dx 0 rmoveto
			30: # dy1 dx2 dy2 dx3 vhcurveto #
				0 dy1 dx2 dy2 dx3 0 rcurveto
			31: # dx1 dx2 dy2 dy3 hvcurveto #
				dx1 0 dx2 dy2 0 dy3 rcurveto
			12 0: - dotsection #
			12 1: # x0 dx0 x1 dx1 x2 dx2 vstem3 #
			12 2: # y0 dy0 y1 dy1 y2 dy2 hstem3 #
			12 6: # asb adx ady bchar achar seac #
			12 7: # sbx sby wx wy sbw #
				sets side and currentpoint to (sbx, sby)
				and charwidth to (wx, wy)
			12 12: p q div <the quotient>
				-> p/q
			12 16: arg1 ... argn n othersubr# callothersubr -
			12 17: - pop <number>
			12 33: # x y setcurrentpoint #

Many indices do not refer to any comand at all;
I believe Adobe uses some of these for private use,
and left space for future aditions.
Since python has no switchg statement, I use
an array to assign names to commands, and put those ames
in a dictionary to assign procedures.  The basic output
is an ordinary PS path made up of moves, lines, curves, closures.

"""

# to get output consistent with Canvases

from Cmd import *

Cmd = [
	MOVETO,
	LINETO,
	CURVETO,
	CLOSEPATH
]

class BuildChar:

	# Here is the initial array of names:

	CmdString = [
		".notdef",
		"hstem",
		".notdef",
		"vstem",
		"vmoveto", # 4

		"rlineto",
		"hlineto",
		"vlineto",
		"rrcurveto",
		"closepath", # 9

		"callsubr",
		"return",
		"escape",
		"hsbw",
		"endchar", # 14

		".notdef",
		".notdef",
		".notdef",
		".notdef",
		".notdef", # 19

		".notdef",
		"rmoveto",
		"hmoveto",
		".notdef",
		".notdef", # 24

		".notdef",
		".notdef",
		".notdef",
		".notdef",
		".notdef", # 29

		"vhcurveto",
		"hvcurveto",
		"dotsection",
		"vstem3",
		"hstem3", # 34

		".notdef",
		".notdef",
		".notdef",
		"seac",
		"sbw", # 39

		".notdef",
		".notdef",
		".notdef",
		".notdef",
		"div", # 44

		".notdef",
		".notdef",
		".notdef",
		"callothersubr",
		"pop", # 49

		".notdef",
		".notdef",
		".notdef",
		".notdef",
		".notdef", # 54

		".notdef",
		".notdef",
		".notdef",
		".notdef",
		".notdef", # 59

		".notdef",
		".notdef",
		".notdef",
		".notdef",
		".notdef", # 64

		"setcurrentpoint" 
	]

	# =============================================

	# y dy
	def hstem(bc):
		bc.opstack = []

	# x dx
	def vstem(bc):
		bc.opstack = []

	# dy
	def vmoveto(bc):
		# get dy from bc's opstack
		dy = bc.opstack[0]
		bc.opstack = []
		bc.cy += dy
		bc.a.append(Cmd[0])
		bc.a.append(bc.cx)
		bc.a.append(bc.cy)

	# dx dy
	def rlineto(bc):
		dx = bc.opstack[0]
		dy = bc.opstack[1]
		bc.opstack = []
		bc.cx += dx
		bc.cy += dy
		bc.a.append(Cmd[1])
		bc.a.append(bc.cx)
		bc.a.append(bc.cy)

	# dx
	def hlineto(bc):
		dx = bc.opstack[0]
		bc.opstack = []
		bc.cx += dx
		bc.a.append(Cmd[1])
		bc.a.append(bc.cx)
		bc.a.append(bc.cy)

	# dy
	def vlineto(bc):
		dy = bc.opstack[0]
		bc.opstack = []
		bc.cy += dy
		bc.a.append(Cmd[1])
		bc.a.append(bc.cx)
		bc.a.append(bc.cy)

	def rrcurveto(bc):
		x1 = bc.cx + bc.opstack[0]
		y1 = bc.cy + bc.opstack[1]
		x2 = x1 + bc.opstack[2]
		y2 = y1 + bc.opstack[3]
		x3 = x2 + bc.opstack[4]
		y3 = y2 + bc.opstack[5]
		bc.opstack = []
		bc.a.append(Cmd[2])
		bc.a.append(x1)
		bc.a.append(y1)
		bc.a.append(x2)
		bc.a.append(y2)
		bc.a.append(x3)
		bc.a.append(y3)
		bc.cx = x3
		bc.cy = y3

	def closepath(bc):
		bc.a.append(Cmd[3])

	# on the assumption that the subr is concerned with stems etc.
	def callsubr(bc):
		#print "callsubr"
		#print "stack = " + str(bc.opstack)
		subr = bc.opstack.pop(-1)
		cs = bc.cs.subroutine( subr )
		p = proc( cs.decryptedString() )
		bc.ht += 1
		bc.procstack.append(p)
		bc.cp = p

	def subreturn(bc):
		# print "returning"
		bc.procstack.pop(-1)
		bc.ht -= 1
		# print "and now = " + str(bc.ht)
		bc.cp = bc.procstack[bc.ht]

	# sbx wx #
	def hsbw(bc):
		bc.sbx = bc.opstack[0]
		bc.sby = 0
		bc.cx = bc.sbx
		bc.cy = 0
		bc.wx = bc.opstack[1]
		bc.wy = 0
		bc.opstack = []

	def endchar(bc):
		bc.ht = -1

	# dx dy
	def rmoveto(bc):
		dx = bc.opstack[0]
		dy = bc.opstack[1]
		bc.opstack = []
		bc.cx += dx
		bc.cy += dy
		bc.a.append(Cmd[0])
		bc.a.append(bc.cx)
		bc.a.append(bc.cy)

	# dx
	def hmoveto(bc):
		dx = bc.opstack[0]
		bc.opstack = []
		bc.cx += dx
		bc.a.append(Cmd[0])
		bc.a.append(bc.cx)
		bc.a.append(bc.cy)

	# 
	def vhcurveto(bc):
		x1 = bc.cx
		y1 = bc.cy + bc.opstack[0]
		x2 = x1 + bc.opstack[1]
		y2 = y1 + bc.opstack[2]
		x3 = x2 + bc.opstack[3]
		y3 = y2
		bc.opstack = []
		bc.a.append(Cmd[2])
		bc.a.append(x1)
		bc.a.append(y1)
		bc.a.append(x2)
		bc.a.append(y2)
		bc.a.append(x3)
		bc.a.append(y3)
		bc.cx = x3
		bc.cy = y3

	def hvcurveto(bc):
		x1 = bc.cx + bc.opstack[0]
		y1 = bc.cy
		x2 = x1 + bc.opstack[1]
		y2 = y1 + bc.opstack[2]
		x3 = x2
		y3 = y2 + bc.opstack[3]
		bc.opstack = []
		bc.a.append(Cmd[2])
		bc.a.append(x1)
		bc.a.append(y1)
		bc.a.append(x2)
		bc.a.append(y2)
		bc.a.append(x3)
		bc.a.append(y3)
		bc.cx = x3
		bc.cy = y3

	def dotsection(bc):
		pass

	def vstem3(bc):
		bc.opstack = []

	def hstem3(bc):
		bc.opstack = []

	def seac(bc):
		pass

	# sbx sby wx wy #
	def sbw(bc):
		bc.sbx = bc.opstack[0]
		bc.sby = bc.opstack[1]
		bc.cx = bc.sbx
		bc.cy = bc.sby
		bc.wx = bc.opstack[2]
		bc.wy = bc.opstack[3]
		bc.opstack = []

	def div(bc):
		q = bc.opstack[-2]*1.0/bc.opstack[-1]
		bc.opstack.pop(-1)
		bc.opstack.pop(-1)
		bc.opstack.append(q)

	def callothersubr(bc):
		# print "callothersubr"
		# print bc.opstack
		subr = bc.opstack[-1]
		bc.opstack.pop(-1)
		n = bc.opstack[-1]
		bc.opstack.pop(-1)
		for i in range(n):
			# delete n items on bc.opstack
			bc.opstack.pop(-1)

	def pop(bc):
		bc.opstack.append(3)

	def setcurrentpoint(bc):
		bc.cx = bc.opstack[0]
		bc.cy = bc.opstack[1]
		bc.opstack = []

	CmdDict= { 
		"hstem":hstem,
		"vstem":vstem,
		"vmoveto":vmoveto,
		"rlineto":rlineto,
		"hlineto":hlineto,

		"vlineto":vlineto,
		"rrcurveto":rrcurveto,
		"closepath":closepath,
		"callsubr":callsubr,
		"return":subreturn,

		"hsbw":hsbw,
		"endchar":endchar,
		"rmoveto":rmoveto,
		"hmoveto":hmoveto,
		"vhcurveto":vhcurveto,

		"hvcurveto":hvcurveto,
		"dotsection":dotsection,
		"vstem3":vstem3,
		"hstem3":hstem3,
		"seac":seac,

		"sbw":sbw,
		"div":div,
		"callothersubr":callothersubr,
		"pop":pop,
		"setcurrentpoint":setcurrentpoint
	}

	# =============================================

	# this turns the decrypted character string s into a special kind of array
	# cmds = moveto, lineto, curveto, closepath in [0, 4), followed by arguments
	# b = the BinarySegment, holding (1) the value of lenIV and (2) the subroutine array
	# necessary to decode these
	def __init__(self, charString ):
		self.cs = charString
		self.cp = proc( charString.decryptedString() )
		self.procstack = [self.cp]
		self.ht = 0
		# builds command array via CmdDict
		self.opstack = []
		self.a = []
		while self.ht > -1:
			cp = self.cp
			c = ord(cp.read())
			# print " * = " + str(c)
			if 0 <= c and c < 32: # a command
				if c == 12:
					c = 32 + ord(cp.read())
				# print "cmd = " + self.CmdString[c]
				self.CmdDict[self.CmdString[c]](self)
			elif 32 <= c and c < 247:
				n = c - 139
				self.opstack.append(n)
			elif 247 <= c and c < 251:
				n = (c-247)*256 + ord(cp.read()) + 108
				self.opstack.append(n)
			elif 251 <= c and c < 255:
				n = -(c-251)*256 - ord(cp.read()) - 108
				self.opstack.append(n)
			else: # c = 255
				n = ord(cp.read())
				for j in range(0, 3):
					n = (n << 8) + ord(cp.read())
				if n >= (1 << 31):
					n -= (1 << 32)
				self.opstack.append(n)
			self.path = self.a # this is a very limited command array of 4 types MOVETO etc.

	PSOp = [ "moveto", "lineto", "curveto", "closepath" ]

	def toPS(self):
		cmd = self.a
		s = ""
		i = 0
		while i < len(cmd):
			c = cmd[i]
			# print "c = " + str(c)
			i += 1
			if c == 0 or c == 1:
				s += " " + str(cmd[i]) + " " + str(cmd[i+1])
				i += 2
			elif c == 2:
				s += " " + str(cmd[i]) + " " + str(cmd[i+1])
				s += " " + str(cmd[i+2]) + " " + str(cmd[i+3])
				s += " " + str(cmd[i+4]) + " " + str(cmd[i+5])
				i += 6
			else: # c == 3
				pass
			s += " " + self.PSOp[c] + "\n"
		return(s)
		
	

		
			
		
		

	
	
