"""
import piscript.VectorUtils as VU
from piscript.Vectors import Vector
import piscript.MatrixUtils as MU
from Fstr import fstr
"""

from PiScript3d import *
import math
import Bezier

class Face:
	def __init__(self, *args):
		if isinstance(args[0], Face):
			f = args[0]
			self.p = f.p
			self.fill = f.fill
			self.nf = f.nf
			self.shading = f.shading
			return
		if len(args) == 1: # polygon
			self.p = args[0]
			self.fill = None
		else: # polygon, color
			self.p = args[0]
			self.fill = args[1]
		self.stroke = None
		self.nf = self.normalfunction(self.p)
		self.shading = PiScript3d.default_shading
		# extras?
		self.extras = []
		for i in range(2,len(args)):
			extras.append(args[i])

	# y = Bernstein array
	def setshading(self, y):
		self.shading = y

	# light, normal
	def shade_factor(self, s):
		s = (s+1)/2.0
		return(Bezier.bernstein(self.shading, s))

	def setnormal(self, nf):
		self.f[1] = nf

	def setfill(self, c):
		self.fill = c

	def setstroke(self, c):
		self.stroke = c

	def reversed(self):
		f = Face(self)
		f.nf = [ -f.nf[0], -f.nf[1], -f.nf[2], -f.nf[3] ]
		return(f)

	def isvisible(self, e):
		s = 0
		for i in range(4):
			s += e[i]*self.nf[i]
		return(s >= 0)

	def is_visible(self, e):
		return(self.isvisible(e))

	def paint(self, ps, *args):
		if len(args) == 1:
			c = args[0]
		else:
			c = self.fill
		stroke = self.stroke
		if c or stroke:
			p = self.p
			n = len(p)
			ps.newpath()
			ps.moveto3d(p[-1][0], p[-1][1], p[-1][2])
			for i in range (n):
				ps.lineto3d(p[i][0], p[i][1], p[i][2])
			ps.closepath3d()
			if c:
				L = ps.get_light(); nf = self.nf
				s = L[0]*nf[0]+L[1]*nf[1]+L[2]*nf[2]
				s = (s + 1.0)/2
				s = Bezier.bernstein(self.shading, s)
				ps.fill(s*c[0], s*c[1], s*c[2])
			if self.stroke:
				ps.stroke(self.stroke)

	def display(self):
		print "face: ", self.p, self.nf

	# --- static utilities ------------------------------------------
		
	def normalfunction(p):
		u = p[0]; v = p[1]; w = p[2]
		a = [ v[0]-u[0], v[1]-u[1], v[2]-u[2] ]
		b = [ w[0]-v[0], w[1]-v[1], w[2]-v[2] ]
		# print a, b
		c = VU.x(a, b)
		r = math.sqrt(c[0]*c[0]+c[1]*c[1]+c[2]*c[2])
		c[0] /= r; c[1] /= r; c[2] /= r
		return [ c[0], c[1], c[2], -(c[0]*u[0]+c[1]*u[1]+c[2]*u[2]) ]
	normalfunction=staticmethod(normalfunction)

	def reverse_array(p):
		n = len(p)
		q = []
		for i in range(n):
			q.append(p[n-i-1])
		return(q)
	reverse_array = staticmethod(reverse_array)

