

"""
from piscript.PiScript import *
import piscript.VectorUtils as VU
from piscript.Vectors import Vector
import piscript.MatrixUtils as MU """

from PiScript import *
import VectorUtils as VU
from Vectors import Vector
import MatrixUtils as MU
from Fstr import fstr

import Bezier

import math

class Face:
	def __init__(self, *args):
		if isinstance(args[0], Face):
			f = args[0]
			self.p = f.p
			self.fill = f.fill
			self.nf = f.nf
			self.shading = f.shading
			return
		if len(args) == 1: # polygon
			self.p = args[0]
			self.fill = None
		else: # polygon, color
			self.p = args[0]
			self.fill = args[1]
		self.stroke = None
		self.nf = self.normalfunction(self.p)
		self.shading = PiScript3d.default_shading
		# extras?
		self.extras = []
		for i in range(2,len(args)):
			extras.append(args[i])

	# y = Bernstein array
	def setshading(self, y):
		self.shading = y

	# light, normal
	def shade_factor(self, s):
		s = (s+1)/2.0
		return(Bezier.bernstein(self.shading, s))

	def setnormal(self, nf):
		self.f[1] = nf

	def setfill(self, c):
		self.fill = c

	def setstroke(self, c):
		self.stroke = c

	def reversed(self):
		f = Face(self)
		f.nf = [ -f.nf[0], -f.nf[1], -f.nf[2], -f.nf[3] ]
		return(f)

	def isvisible(self, e):
		s = 0
		for i in range(4):
			s += e[i]*self.nf[i]
		return(s >= 0)

	def is_visible(self, e):
		return(self.isvisible(e))

	def paint(self, ps, *args):
		if len(args) == 1:
			c = args[0]
		else:
			c = self.fill
		stroke = self.stroke
		if c or stroke:
			p = self.p
			n = len(p)
			ps.newpath()
			ps.moveto3d(p[-1][0], p[-1][1], p[-1][2])
			for i in range (n):
				ps.lineto3d(p[i][0], p[i][1], p[i][2])
			ps.closepath3d()
			if c:
				L = ps.get_light(); nf = self.nf
				s = L[0]*nf[0]+L[1]*nf[1]+L[2]*nf[2]
				s = (s + 1.0)/2
				s = Bezier.bernstein(self.shading, s)
				ps.fill(s*c[0], s*c[1], s*c[2])
			if self.stroke:
				ps.stroke(self.stroke)

	def display(self):
		print "face: ", self.p, self.nf

	# --- static utilities ------------------------------------------
		
	def normalfunction(p):
		u = p[0]; v = p[1]; w = p[2]
		a = [ v[0]-u[0], v[1]-u[1], v[2]-u[2] ]
		b = [ w[0]-v[0], w[1]-v[1], w[2]-v[2] ]
		# print a, b
		c = VU.x(a, b)
		r = math.sqrt(c[0]*c[0]+c[1]*c[1]+c[2]*c[2])
		c[0] /= r; c[1] /= r; c[2] /= r
		return [ c[0], c[1], c[2], -(c[0]*u[0]+c[1]*u[1]+c[2]*u[2]) ]
	normalfunction=staticmethod(normalfunction)

	def reverse_array(p):
		n = len(p)
		q = []
		for i in range(n):
			q.append(p[n-i-1])
		return(q)
	reverse_array = staticmethod(reverse_array)

class SmoothConvexSurface:
	# f = array of triangular faces with normal vectors
	def __init__(self, f):
		self.f = f

	# ps = Vector				
	def paint(self, ps):
		e = ps.get_eye()
		l = ps.get_light()
		C = self.color
		n = len(self.f)
		MAX = 1024
		# print n, " triangles"
		while (n > 0):
			N = MAX
			if N > n:
				N = n
			n -= N
			data_source = []
			# data source = an array of triangles [ P Q R ]
			# each triangle is an array of three `enhanced' vertices [ 0 x y r g b ]
			"""
			s = "\n"
			s += "newpath\n"
			s += "<<\n"
			s += "/ShadingType 4\n"
  			s += "/ColorSpace [ /DeviceRGB ]\n"
  			s += "/DataSource [\n"
			"""
			for i in range(N):
				F = self.f[i]
				if F.is_visible(e):
					t = F.p # each face is a triangle
					a = t[0]; b = t[1]; c = t[2]
					a = [ a[0], a[1], a[2], 1 ]
					b = [ b[0], b[1], b[2], 1 ]
					c = [ c[0], c[1], c[2], 1 ]
					# add shfill VU lines to s
					sh = VU.mul(l, a)
					sh = F.shade_factor(sh)
					u = ps.transform2d(a)
					T = []
					T.append([ (u[0], u[1]), (sh*C[0], sh*C[1], sh*C[2])])
					sh = VU.mul(l, b)
					sh = F.shade_factor(sh)
					u = ps.transform2d(b)
					T.append([ (u[0], u[1]), (sh*C[0], sh*C[1], sh*C[2])])
					sh = VU.mul(l, c)
					sh = F.shade_factor(sh)
					u = ps.transform2d(c)
					T.append([ (u[0], u[1]), (sh*C[0], sh*C[1], sh*C[2])])
					data_source.append(T)
			"""
			s += "]\n"
			s += ">>\n"
			s += "shfill\n\n"
			ps.page += s
			"""
			ps.shfill(data_source)

class ConvexSurface:
	# F = array of faces
	def __init__(self, F):
		self.faces = F	

	def paint(self, ps):
		for i in range(len(self.faces)):
			if self.faces[i].is_visible(ps.get_eye()):
				self.faces[i].paint(ps)

	def setshading(self, y):
		f = self.faces
		for F in f:
			F.setshading(y)

	def setfill(self, c):
		f = self.faces
		for F in f:
			F.setfill(c)

	def setstroke(self, c):
		f = self.faces
		for F in f:
			F.setstroke(c)



# ---------------------------------------------------------------------------
# ---------------------------------------------------------------------------
# ---------------------------------------------------------------------------



class PiScript3d(PiScript):

	default_shading = [ 0.3, 0.5, 0.9, 1.0 ]

	def __init__(self, device, *args):
		# print "device =", device
		PiScript.__init__(self, device, *args)
		""" self.setlinecap(1)
		self.setlinejoin(1) """
		self.gstack3d = [
			[[[1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,1]], # CTM
			[[1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,1]], # inverse of CTM
			[[1,0,0,0],[0,1,0,0],[0,0,0,1]]] # display matrix
		]
		L = [ -0.5, 1, 0.25, 0 ]
		self.light = L
		self.set_eye([ 0, 0, 1, 0 ])
		self.setlinecap(1)
		self.setlinejoin(1)
		self.lm = None
		# self.cpt also?  
		# It is not used, except potentially to tell if there is an active 3d path
		# self.lm can be used for that

	# --- graphics state and coordinate changes ------------------------

	def gsave3d(self):
		g = self.gstack3d[-1]
		self.gstack3d.append([MU.clone(g[0]), MU.clone(g[1]), g[2]])

	def grestore3d(self):
		self.gstack3d.pop(-1)

	def set_eye(self, e): 
		self.eye = e
		self.gstack3d[-1][2] = [
			[e[2], 0, -e[0], 0],
			[0, e[2], -e[1], 0],
			[0, 0, -e[3], e[2]],
		]
		g = self.gstack3d[-1]

	# returns virtual eye
	def get_eye(self):
		g = self.gstack3d[-1]
		t = g[1]
		# print Vector.prod(g[0], g[1])
		e = self.eye
		v = []
		# return t*e
		for i in range(4):
			s = 0
			for j in range(4):
				s += t[i][j]*e[j]
			v.append(s)
		return(v)

	def homogeneous2d(self, x):
		t = self.gstack3d[-1][2] # t = display matrix
		y = []
		for i in range(3):
			s = 0
			for j in range(4):
				s += t[i][j]*x[j]
			y.append(s)
		return(y)

	def display(self, x):
		t = self.gstack3d[-1][2] # t = display matrix
		y = []
		for i in range(3):
			s = 0
			for j in range(4):
				s += t[i][j]*x[j]
			y.append(s)
		return([y[0]*1.0/y[2], y[1]*1.0/y[2]])

	def scale3d(self, *args):
		if len(args) == 3:
			x = args[0]
			y = args[1]
			z = args[2]
		else:
			x = args[0][0]
			y = args[0][1]
			z = args[0][2]
		t = self.gstack3d[-1][0]
		T = self.gstack3d[-1][1]
		for i in range(4):
			t[i][0] *= x
			t[i][1] *= y
			t[i][2] *= z
		for j in range(4):
			T[0][j] /= x*1.0
			T[1][j] /= y*1.0
			T[2][j] /= z*1.0

	"""
		[ * * * t[0][3]
		  * * * t[1][3]
		  * * * t[2][3]
		  * * * t[3][3] ]
		[ 1 0 0 x
		  0 1 0 y
		  0 0 1 z
		  0 0 0 1 ]
	"""
	def translate3d(self, *args):
		if len(args) == 3:
			x = args[0]
			y = args[1]
			z = args[2]
		else:
			x = args[0][0]
			y = args[0][1]
			z = args[0][2]
		t = self.gstack3d[-1][0]
		T = self.gstack3d[-1][1]
		for i in range(4):
			t[i][3] += x*t[i][0] + y*t[i][1] + z*t[i][2]
		for j in range(4):
			T[0][j] -= x*T[3][j]
			T[1][j] -= y*T[3][j]
			T[2][j] -= z*T[3][j]

	# t*r, r^{-1}*T
	def rotate3d(self, a, A):
		r = self.rotationmatrix(a, A)
		t = self.gstack3d[-1][0]
		T = self.gstack3d[-1][1]
		p = []
		for i in range(4):
			q = []
			for k in range(3):
				s = 0
				for j in range(3):
					s += t[i][j]*r[j][k]
				q.append(s)
			p.append(q)
		# p = t*r
		for i in range(4): # i-th row of t
			for k in range(3): # k-th column of r
				t[i][k] = p[i][k]

		p = []
		for i in range(3):
			q = []
			for k in range(4):
				s = 0
				for j in range(3):
					s += r[j][i]*T[j][k]
				q.append(s)
			p.append(q)
		# p = r^{-1}T
		for i in range(3): # we (anti-)rotate first three rows
			T[i] = p[i]

	# t*F, None
	"""
		[ t00 t01 t02 t03 
		  t10 t11 t12 t13
		  t20 t21 t22 t23
		  t30 t31 t32 t33 ]
	#
		[ -f(P)+Ax_{P} 	& -Bx_{P}		& Cx_{P} 		& Dx_{P} 
		Ay_{P} 		& -f(P)+By_{P}	& Cy_{P} 		& Dy_{P} 
		Az_{P} 		& Bz_{P}		& -f(Q)+Cz_{P} 	& Dz_{P} 
		A 			& B 		& C 		& -f(Q)+D ]
	"""
	def project(self, F, P):
		fP = F[0]*P[0]+F[1]*P[1]+F[2]*P[2]+F[3]*P[3]
		F = [
			[ -fP+F[0]*P[0] ,  F[1]*P[0]	,  F[2]*P[0] 	,  F[3]*P[0] ],
			[  F[0]*P[1] 	, -fP+F[1]*P[1]	, -F[2]*P[1] 	,  F[3]*P[1] ], 
			[  F[0]*P[2]	,  F[1]*P[2]	, -fP+F[2]*P[2]	,  F[3]*P[2]  ], 
			[  F[0]*P[3] 	,  F[1]*P[3] 	,  F[2]*P[3] 	, -fP+F[3]*P[3] ],
		]
		t = self.gstack3d[-1][0]
		self.gstack3d[-1][1] = None

		p = []
		for i in range(4):
			q = []
			for k in range(4):
				s = 0
				for j in range(4):
					s += t[i][j]*F[j][k]
				q.append(s)
			p.append(q)
		# p = t*F
		for i in range(4): # i-th row of t
			for k in range(4): # k-th column of r
				t[i][k] = p[i][k]
		
	# --- drawing ---------------------------------------

	def transform3d(self, v):
		return(MU.transform(self.gstack3d[-1][0], v))

	def transform2d(self, v):
		v = MU.transform(self.gstack3d[-1][0], v)
		return self.display(v)
		
	def moveto3d(self, *args):
		if len(args) == 3:
			x = args[0]
			y = args[1]
			z = args[2]
		else:
			x = args[0][0]
			y = args[0][1]
			z = args[0][2]
		v = [ x, y, z, 1 ]
		v = MU.transform(self.gstack3d[-1][0], v)
		self.cpt = v
		self.lm = v
		V = self.display(v)
		self.moveto(V[0], V[1])

	def rmoveto3d(self, *args):
		if len(args) == 3:
			dx = args[0]
			dy = args[1]
			dz = args[2]
		else:
			dx = args[0][0]
			dy = args[0][1]
			dz = args[0][2]
		cpt = self.cpt
		dv = [ dx, dy, dz, 0 ]
		dv = MU.transform(self.gstack3d[-1][0], dv)
		v = [ cpt[0]+dv[0], cpt[1]+dv[1], cpt[2]+dv[2], 1]
		self.cpt = v
		V = self.display(v)
		self.moveto(V[0], V[1])

	def lineto3d(self, *args):
		if len(args) == 3:
			x = args[0]
			y = args[1]
			z = args[2]
		else:
			x = args[0][0]
			y = args[0][1]
			z = args[0][2]
		v = [ x, y, z, 1 ]
		v = MU.transform(self.gstack3d[-1][0], v)
		self.cpt = v
		V = self.display(v)
		self.lineto(V[0],V[1])

	def rlineto3d(self, *args):
		if len(args) == 3:
			dx = args[0]
			dy = args[1]
			dz = args[2]
		else:
			dx = args[0][0]
			dy = args[0][1]
			dz = args[0][2]
		""" OLD:
		cpt = MU.transform(self.gstack3d[-1][1], self.cpt)
		v = [ cpt[0]+dx, cpt[1]+dy, cpt[2]+dz, cpt[3] ]
		v = MU.transform(self.gstack3d[-1][0], v)
		self.cpt = v """
		cpt = self.cpt
		dv = [ dx, dy, dz, 0 ]
		dv = MU.transform(self.gstack3d[-1][0], dv)
		v = [ cpt[0]+dv[0], cpt[1]+dv[1], cpt[2]+dv[2], 1]
		self.cpt = v
		V = self.display(v)
		self.lineto(V[0], V[1])

	# expected to be planar?
	def curveto3d(self, P1, P2, P3):
		cpt = self.cpt
		p0 = self.homogeneous2d([cpt[0], cpt[1], cpt[2], 1])
		p1 = MU.transform(self.gstack3d[-1][0], [ P1[0], P1[1], P1[2], 1 ])
		p1 = self.homogeneous2d(p1)
		w = 1.0*p0[2]
		c = p1[2]/w - 1
		x0 = p0[0]/w
		x1 = p1[0]/w
		y0 = p0[1]/w
		y1 = p1[1]/w
		x1 -= c*x0
		y1 -= c*y0
		p2 = MU.transform(self.gstack3d[-1][0], [ P2[0], P2[1], P2[2], 1 ])
		p2 = self.homogeneous2d(p2)
		p3 = MU.transform(self.gstack3d[-1][0], [ P3[0], P3[1], P3[2], 1 ])
		self.cpt = p3
		p3 = self.homogeneous2d(p3)
		w = 1.0*p3[2]
		c = p2[2]/w - 1
		x3 = p3[0]/w
		y3 = p3[1]/w
		x2 = p2[0]/w
		y2 = p2[1]/w
		x2 -= c*x3
		y2 -= c*y3
		self.curveto(x1,y1,x2,y2,x3,y3)

	def closepath3d(self):
		v = MU.transform(self.gstack3d[-1][0], self.lm)
		self.cpt = v
		self.closepath()

	# ------------------------------------------------------------------

	def Xmoveto(self, *args):
		if len(args) == 2 or len(args) == 1 and len(args[0]) == 2:
			return PiScript.moveto(self, *args)
		if len(args) == 3:
			x = args[0]
			y = args[1]
			z = args[2]
		else:
			x = args[0][0]
			y = args[0][1]
			z = args[0][2]
		v = [ x, y, z, 1 ]
		v = MU.transform(self.gstack3d[-1][0], v)
		self.cpt = v
		self.lm = v
		V = self.display(v)
		self.moveto(V[0], V[1])

	def Xrmoveto(self, *args):
		if len(args) == 2 or len(args) == 1 and len(args[0]) == 2:
			return PiScript.rmoveto(self, *args)
		if len(args) == 3:
			dx = args[0]
			dy = args[1]
			dz = args[2]
		else:
			dx = args[0][0]
			dy = args[0][1]
			dz = args[0][2]
		cpt = self.cpt
		dv = [ dx, dy, dz, 0 ]
		dv = MU.transform(self.gstack3d[-1][0], dv)
		v = [ cpt[0]+dv[0], cpt[1]+dv[1], cpt[2]+dv[2], 1]
		self.cpt = v
		V = self.display(v)
		self.moveto(V[0], V[1])

	def Xlineto(self, *args):
		if len(args) == 2 or len(args) == 1 and len(args[0]) == 2:
			return PiScript.lineto(self, *args)
		if len(args) == 3:
			x = args[0]
			y = args[1]
			z = args[2]
		else:
			x = args[0][0]
			y = args[0][1]
			z = args[0][2]
		v = [ x, y, z, 1 ]
		v = MU.transform(self.gstack3d[-1][0], v)
		self.cpt = v
		V = self.display(v)
		self.lineto(V[0],V[1])

	def Xrlineto3d(self, *args):
		if len(args) == 2 or len(args) == 1 and len(args[0]) == 2:
			return PiScript.rlineto(self, *args)
		if len(args) == 3:
			dx = args[0]
			dy = args[1]
			dz = args[2]
		else:
			dx = args[0][0]
			dy = args[0][1]
			dz = args[0][2]
		""" OLD:
		cpt = MU.transform(self.gstack3d[-1][1], self.cpt)
		v = [ cpt[0]+dx, cpt[1]+dy, cpt[2]+dz, cpt[3] ]
		v = MU.transform(self.gstack3d[-1][0], v)
		self.cpt = v """
		cpt = self.cpt
		dv = [ dx, dy, dz, 0 ]
		dv = MU.transform(self.gstack3d[-1][0], dv)
		v = [ cpt[0]+dv[0], cpt[1]+dv[1], cpt[2]+dv[2], 1]
		self.cpt = v
		V = self.display(v)
		self.lineto(V[0], V[1])

	# (P1, P2) or (P1, P2, P3)
	def Xcurveto(self, *args):
		if len(args) == 2:
			return PiScript.lineto(self, *args)
		cpt = self.cpt
		p0 = self.homogeneous2d([cpt[0], cpt[1], cpt[2], 1])
		p1 = MU.transform(self.gstack3d[-1][0], [ P1[0], P1[1], P1[2], 1 ])
		p1 = self.homogeneous2d(p1)
		w = 1.0*p0[2]
		c = p1[2]/w - 1
		x0 = p0[0]/w
		x1 = p1[0]/w
		y0 = p0[1]/w
		y1 = p1[1]/w
		x1 -= c*x0
		y1 -= c*y0
		p2 = MU.transform(self.gstack3d[-1][0], [ P2[0], P2[1], P2[2], 1 ])
		p2 = self.homogeneous2d(p2)
		p3 = MU.transform(self.gstack3d[-1][0], [ P3[0], P3[1], P3[2], 1 ])
		self.cpt = p3
		p3 = self.homogeneous2d(p3)
		w = 1.0*p3[2]
		c = p2[2]/w - 1
		x3 = p3[0]/w
		y3 = p3[1]/w
		x2 = p2[0]/w
		y2 = p2[1]/w
		x2 -= c*x3
		y2 -= c*y3
		self.curveto(x1,y1,x2,y2,x3,y3)

	def Xclosepath(self):
		# if a 3d path active ...
		if self.lm:
			v = MU.transform(self.gstack3d[-1][0], self.lm)
			self.cpt = v
		PiScript.closepath(self)

	def Xstroke(self, *args):
		self.lm = None
		PiScript.stroke(self, *args)

	def Xfill(self, *args):
		self.lm = None
		PiScript.fill(self, *args)

	# --- surfaces and shading -----------------------------------------

	def set_light(self, L): 
		self.light = L

	# returns virtual light
	def get_light(self):
		t = self.gstack3d[-1][1]
		# print str(t)
		e = self.light
		v = []
		# return t*e
		for i in range(4):
			s = 0
			for j in range(4):
				s += t[i][j]*e[j]
			v.append(s)
		r = math.sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2])
		v[0] /= r; v[1] /= r; v[2] /= r
		return(v)

	import math

	# u rotated by A around axis a
	def rotated(self, u, a, A):
		r = VU.length(a)*1.0
		for i in range(3):
			a[i] /= r
		d = VU.mul(u, a)
		u0 = [ a[0]*d, a[1]*d, a[2]*d ] # projection of u onto axis
		u1 = [ u[0] - u0[0], u[1] - u0[1], u[2] - u0[2] ]
		# projection of u onto plane perpendicular to axis 
		u2 = VU.x(a, u1)
		A *= self.toRad
		c = math.cos(A)
		s = math.sin(A)
		v = []
		for i in range(3):
			v.append(u0[i] + c*u1[i] + s*u2[i])
		return(v)

	# matrices are arrays of rows
	def rotationmatrix(self, axis, A):
		a = self.rotated([1,0,0], axis, A)
		b = self.rotated([0,1,0], axis, A)
		c = self.rotated([0,0,1], axis, A)
		R = [
			[ a[0], b[0], c[0] ],
			[ a[1], b[1], c[1] ],
			[ a[2], b[2], c[2] ],
		]
		return R



