#include "kpathsea/kpathsea.h"
#include "Python.h"

// Flag to determine if we managed to initialize the kpathsea library correctly.
static boolean isFunctional = FALSE;

static PyObject *pykpse_find( PyObject *self, PyObject *args )
{
	if( !isFunctional )
	{
		PyErr_SetString( PyExc_RuntimeError, "kpathsea library not initialized" );
		return NULL;
	}

	char *fName = NULL;
	int typeIndex;
	if( !PyArg_ParseTuple( args, "si", &fName, &typeIndex ) )
	{
		return NULL;
	}
	
	char *found = kpse_find_file (fName, typeIndex, false);
	if( found == NULL )
	{
		Py_RETURN_NONE;
	}
	
	PyObject *rv = PyString_FromString( found );
	free( found );
	return rv;	
}

PyMethodDef methods[] = {
    {"find", pykpse_find, METH_VARARGS, "Returns a the location of a TeX file."},
//    {"initialize", pykpse_initialize, METH_VARARGS, "Initializes the KPATHSEA library."},
    {NULL, NULL, 0, NULL}
};

PyMODINIT_FUNC 
initpykpse()
{
	kpse_set_program_name( "dvips", "dvips" );	

  PyObject *module = Py_InitModule("pykpse", methods);   

	// Verify that we have a valid 
	string auto_loc = kpse_var_value("SELFAUTOLOC");
	if( auto_loc == NULL )
	{
		PyErr_WarnEx(NULL, "\nUnable to initialize kpathsea library: path to TeX distribution not found.",1);
		(void) PyModule_AddObject( module, "isFunctional", Py_False );
		isFunctional = FALSE;
	}
	else
	{
		(void) PyModule_AddObject( module, "isFunctional", Py_True );
		isFunctional = TRUE;
	}
		
	(void) PyModule_AddIntConstant( module, "TFM_TYPE", kpse_tfm_format );
	(void) PyModule_AddIntConstant( module, "VF_TYPE", kpse_vf_format );
	(void) PyModule_AddIntConstant( module, "TYPE1_TYPE", kpse_type1_format );
	(void) PyModule_AddIntConstant( module, "FONTMAP_TYPE", kpse_fontmap_format );
	(void) PyModule_AddIntConstant( module, "ENC_TYPE", kpse_enc_format );
}
