
import math

def clone(v):
	u = []
	for i in range(len(v)):
		u.append(v[i])
	return u

def minus(u):
	v = []
	n = len(u)
	for i in range(n):
		v.append(-u[i])
	return v

# dot product, scalar product
def mul(u, v):
	if isarray(v):
		n = len(u)
		p = u[0]*v[0]
		for i in range(1, n):
			p += u[i]*v[i]
		return p
	else: # V = scale
		n = len(u)
		V = []
		for i in range(n):
			V.append (u[i]*v)
		return V

# scalar division by c
def div(u, c):
	n = len(u)
	v = []
	for i in range(n):
		v.append (u[i]/c)
	return v

# sum called by +
def sum(u, v):
	n = len(u)
	p = []
	for i in range(0, n):
		p.append(u[i] + v[i])
	return p

# sum called by +
def diff(u, v):
	n = len(u)
	p = []
	for i in range(0, n):
		p.append(u[i] - v[i])
	return p

"""	[ u0 u1 u2 ]
	[ v0 v1 v2 ] """

def x(u, v):
	w = []
	w.append(u[1]*v[2]-u[2]*v[1])
	w.append(u[2]*v[0]-u[0]*v[2])
	w.append(u[0]*v[1]-u[1]*v[0])
	return w

def cross(u, v):
	w = []
	w.append(u[1]*v[2]-u[2]*v[1])
	w.append(u[2]*v[0]-u[0]*v[2])
	w.append(u[0]*v[1]-u[1]*v[0])
	return w

# only for 2D vectors
def arg(u):
	return math.atan2(u[1], u[0])

def perp(v):
	return([-v[1], v[0]])

def rotated(u, a):
	c = math.cos(a); s = math.sin(a)
	return [ c*u[0] - s*u[1], s*u[0] + c*u[1] ]

def reflected(f, u, v): # also in Vectors? f(x) = 0 -> x+u -> x-u
	# grad = (f[0], f[1])
	# f(x+u) = <f, x> + <grad(f), u>
	# f(x) = 0, x+u -> x+u - c<f, x+u> u
	# c = 2/<grad(f), u> 
	r = f[0]*u[0]+f[1]*u[1]
	c = 2*(f[0]*v[0]+f[1]*v[1]+f[2])/float(r)
	return ( v[0]-c*u[0], v[1]-c*u[1])

def length(u):
	max = 0
	for x in u:
		if abs(x) > max:
			max = abs(x)
	if max == 0:
		return 0
	s = 0
	for x in u:
		r = x/max
		s += r*r
	return max*math.sqrt(s)

# normalized so |gradient| = 1
def linethrough(P, Q):
	v = [ Q[0]-P[0], Q[1]-P[1] ]
	r = length(v)
	A = -v[1]/r
	B = v[0]/r
	C = A*P[0]+B*P[1]
	return [ A, B, -C ]

# called by print
def string(u):
	s = "["
	n = len(u)
	for i in range(n):
		s += " " + fstr(u[i])
		if i < n-1:
			s += ","
	s += " ]"
	return(s)

def rotate(u, A):
	c = math.cos(A); s = math.sin(A)
	x = c*u[0]-s*u[1]; y = s*u[0]+c*u[1] 
	u[0] = x; u[1] = y

def interpolated(p, q, t):
	s = 1-t
	r = [ ]
	for i in range(len(p)):
		r.append(s*p[i]+t*q[i])
	return r

def isarray(a):
	if "__getitem__" in dir(a):
		return True
	return False

def distance(P, Q):
	n = len(P)
	v = [ 0 ] *n
	for i in range(n):
		v[i] = Q[i] - P[i]
	return length(v)

